#include <iostream>

#define GLM_FORCE_RADIANS
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>

#include "Car.h"
#include "Particle.h"
#include "Spring.h"
#include "MatrixStack.h"
#include "Program.h"
#include "GLSL.h"

using namespace std;
using namespace Eigen;

#ifndef M_PI
    #define M_PI 3.14159265358979323846
#endif

typedef Eigen::Triplet<double> Td;

shared_ptr<Spring> Car::createSpring(const shared_ptr<Particle> p0, const shared_ptr<Particle> p1, double E)
{
	auto s = make_shared<Spring>(p0, p1);
	s->E = E;
	Vector3d x0 = p0->x;
	Vector3d x1 = p1->x;
	Vector3d dx = x1 - x0;
	s->L = dx.norm();
	return s;
}

void Car::addFaceToElementBuf(int xStart, int xEnd, int yStart, int yEnd, int zStart, int zEnd, int xOffset, int yOffset, int zOffset){
	for(int i = xStart; i < xEnd; i++) {
		for(int j = yStart; j < yEnd; j++) {
			for(int k = zStart; k < zEnd; k++) {
				// Triangle strip
				eleBuf.push_back(i*numY*numZ + j*numZ + k);
				eleBuf.push_back((i+xOffset)*numY*numZ + (j+yOffset)*numZ + (k+zOffset));
			}
		}
	}
}

Car::Car(int numX, int numY, int numZ,
			const Vector3d &center,
			double width, double length, double height,
			double mass,
			double stiffness)
{
	assert(numX > 1);
	assert(numY > 1);
	assert(numZ > 1);
	assert(mass > 0.0);
	assert(stiffness > 0.0);
	
	this->numX = numX;
	this->numY = numY;
	this->numZ = numZ;

	Vector3d start = center - Vector3d(width, height, length)/2;
	Vector3d end = center + Vector3d(width, height, length)/2;
	
	// Create particles
	n = 0;
	double r = 0.02; // Used for collisions
	int nVerts = numX * numY * numZ;
	for(int i = 0; i < numX; i++) {
		for(int j = 0; j < numY; j++) {
			for(int k = 0; k < numZ; k++) {
				Vector3d x = start + Vector3d(i/(numX-1.0) * width, j/(numY-1.0) * height, k/(numZ-1.0) * length);

				auto p = make_shared<Particle>();
				particles.push_back(p);
				p->r = r;
				p->x = x;
				p->v << 0.0, 0.0, 0.0;
				p->m = mass/(nVerts);
				
				p->fixed = false;
				p->i = n;
				n += 3;
			}
		}
	}
	
	for(int i = 0; i < numX; i++) {
		for(int j = 0; j < numY; j++) {
			for(int k = 0; k < numZ; k++) {
				// Create x springs
				if(i < numX-1){
					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + k], 
						particles[(i+1)*numY*numZ + j*numZ + k], 
						stiffness));
				}

				// Create y springs
				if(j < numY-1){
					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + k], 
						particles[i*numY*numZ + (j+1)*numZ + k], 
						stiffness));
				}

				// Create z springs
				if(k < numZ-1){
					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + k], 
						particles[i*numY*numZ + j*numZ + (k+1)], 
						stiffness));
				}

				// Create xy shear springs
				if(i < numX-1 && j < numY-1){
					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + k], 
						particles[(i+1)*numY*numZ + (j+1)*numZ + k], 
						stiffness));

					springs.push_back(createSpring(
						particles[i*numY*numZ + (j+1)*numZ + k], 
						particles[(i+1)*numY*numZ + j*numZ + k], 
						stiffness));
				}

				// Create xz shear springs
				if(i < numX-1 && k < numZ-1){
					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + k], 
						particles[(i+1)*numY*numZ + j*numZ + (k+1)], 
						stiffness));

					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + (k+1)], 
						particles[(i+1)*numY*numZ + j*numZ + k], 
						stiffness));
				}

				// Create yz shear springs
				if(j < numY-1 && k < numZ-1){
					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + k], 
						particles[i*numY*numZ + (j+1)*numZ + (k+1)], 
						stiffness));

					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + (k+1)], 
						particles[i*numY*numZ + (j+1)*numZ + k], 
						stiffness));
				}
				
				double stiffnessBend = stiffness;// * 200;

				// Create x bending springs
				if(i < numX-2){
					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + k], 
						particles[(i+2)*numY*numZ + j*numZ + k], 
						stiffnessBend));
				}

				// Create y bending springs
				if(j < numY-2){
					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + k], 
						particles[i*numY*numZ + (j+2)*numZ + k], 
						stiffnessBend));
				}

				// Create z bending springs
				if(k < numZ-2){
					springs.push_back(createSpring(
						particles[i*numY*numZ + j*numZ + k], 
						particles[i*numY*numZ + j*numZ + (k+2)], 
						stiffnessBend));
				}
			}
		}
	}

	// Build system matrices and vectors
	// M.resize(n,n);
	// K.resize(n,n);
	v.resize(n);
	f.resize(n);
	
	// Build vertex buffers
	posBuf.clear();
	norBuf.clear();
	// texBuf.clear();
	eleBuf.clear();
	posBuf.resize(nVerts*3);
	norBuf.resize(nVerts*3);
	updatePosNor();

	// // Texture coordinates (don't change)
	// for(int i = 0; i < rows; ++i) {
	// 	for(int j = 0; j < cols; ++j) {
	// 		texBuf.push_back(i/(rows-1.0));
	// 		texBuf.push_back(j/(cols-1.0));
	// 	}
	// }

	// Elements (don't change)
	// for(int i = 0; i < rows-1; ++i) {
	// 	for(int j = 0; j < cols; ++j) {
	// 		int k0 = i*cols + j;
	// 		int k1 = k0 + cols;
	// 		// Triangle strip
	// 		eleBuf.push_back(k0);
	// 		eleBuf.push_back(k1);
	// 	}
	// }

	// Elements
	// x left-right, y up-down, z forward-back
	addFaceToElementBuf(0, numX-1, 
						0, 1, 
						0, numZ, 
						1, 0, 0);
	addFaceToElementBuf(0, numX-1, 
						numY-1, numY, 
						0, numZ, 
						1, 0, 0);

	addFaceToElementBuf(0, 1, 
						0, numY-1, 
						0, numZ, 
						0, 1, 0);
	addFaceToElementBuf(numX-1, numX, 
						0, numY-1, 
						0, numZ, 
						0, 1, 0);

	addFaceToElementBuf(0, numX-1, 
						0, numY, 
						0, 1, 
						1, 0, 0);
	addFaceToElementBuf(0, numX-1, 
						0, numY, 
						numZ-1, numZ, 
						1, 0, 0);
}

Car::~Car()
{
}

void Car::load(const string &RESOURCE_DIR) {
	wheelShape = make_shared<Shape>();
	wheelShape->loadMesh(RESOURCE_DIR + "wheel.obj");

	wheels.push_back(make_shared<Wheel>(wheelShape, true));
	wheels.push_back(make_shared<Wheel>(wheelShape, true));
	wheels.push_back(make_shared<Wheel>(wheelShape, false));
	wheels.push_back(make_shared<Wheel>(wheelShape, false));
}

void Car::tare()
{
	for(int k = 0; k < (int)particles.size(); ++k) {
		particles[k]->tare();
	}
}

void Car::reset()
{
	for(int k = 0; k < (int)particles.size(); ++k) {
		particles[k]->reset();
	}
	updatePosNor();
}

void Car::updatePosNor()
{
	// Position
	for(int i = 0; i < numX; i++) {
		for(int j = 0; j < numY; j++) {
			for(int k = 0; k < numZ; k++) {
				int ind = i*numY*numZ + j*numZ + k;
				Vector3d x = particles[ind]->x;
			
				posBuf[3*ind+0] = x(0);
				posBuf[3*ind+1] = x(1);
				posBuf[3*ind+2] = x(2);
			}
		}
	}

	// Normal
	// Each particle has four neighbors
	//
	//      v1
	//     /|\
	// u0 /_|_\ u1
	//    \ | /
	//     \|/
	//      v0
	//
	// Use these four triangles to compute the normal
	for(int i = 0; i < numX; i++) {
		for(int j = 0; j < numY; j++) {
			for(int k = 0; k < numZ; k++) {
				int ind = i*numY*numZ + j*numZ + k;

				Vector3d nor(0.0, 0.0, 0.0);
				Vector3d x = particles[ind]->x;

				Vector3d left;
				Vector3d right;
				Vector3d up;
				Vector3d down;

				if(j == numY-1 || j == 0){ // top or bottom
					int dir = j == numY-1 ? -1 : 1;
					Vector3d sink = particles[i*numY*numZ + (j+dir)*numZ + k]->x;

					left = i > 0 ? particles[(i-1)*numY*numZ + j*numZ + k]->x : sink;
					right = i < numX-1 ? particles[(i+1)*numY*numZ + j*numZ + k]->x : sink;
					up = k > 0 ? particles[i*numY*numZ + j*numZ + (k-1)]->x : sink; // z is +1 towards us, -1 away
					down = k < numZ-1 ? particles[i*numY*numZ + j*numZ + (k+1)]->x : sink; 

					if(j == 0){ // bottom
						Vector3d temp = left;
						left = right;
						right = temp;
					}
				}
				else if(i == 0 || i == numX-1){ // left or right
					int dir = i == numX-1 ? -1 : 1;
					Vector3d sink = particles[(i+dir)*numY*numZ + j*numZ + k]->x;
					
					left = j > 0 ? particles[i*numY*numZ + (j-1)*numZ + k]->x : sink;
					right = j < numY-1 ? particles[i*numY*numZ + (j+1)*numZ + k]->x : sink;
					up = k > 0 ? particles[i*numY*numZ + j*numZ + (k-1)]->x : sink;
					down = k < numZ-1 ? particles[i*numY*numZ + j*numZ + (k+1)]->x : sink;

					if(i == numX-1){ // right
						Vector3d temp = left;
						left = right;
						right = temp;
					}
				}
				else if(k == numZ-1 || k == 0){ // front or back
					int dir = k == numZ-1 ? -1 : 1;
					Vector3d sink = particles[i*numY*numZ + j*numZ + (k+dir)]->x;
					
					left = i < numX-1 ? particles[(i+1)*numY*numZ + j*numZ + k]->x : sink;
					right = i > 0 ? particles[(i-1)*numY*numZ + j*numZ + k]->x : sink;
					up = j > 0 ? particles[i*numY*numZ + (j-1)*numZ + k]->x : sink;
					down = j < numY-1 ? particles[i*numY*numZ + (j+1)*numZ + k]->x : sink;

					if(k == 0){ // back
						Vector3d temp = left;
						left = right;
						right = temp;
					}
				}
				else {
					norBuf[3*ind+0] = 0;
					norBuf[3*ind+1] = 0;
					norBuf[3*ind+2] = 0;
					continue;
				}

				nor += (right-x).cross(up-x).normalized();
				nor += (up-x).cross(left-x).normalized();
				nor += (left-x).cross(down-x).normalized();
				nor += (down-x).cross(right-x).normalized();

				int count = 4;
				if((i == 0 || i == numX-1) && (j == 0 || j == numY-1) && (k == 0 || k == numZ-1))
					count--; // one of the 8 corners with only 3 edges

				nor /= count;
				nor.normalize();
				norBuf[3*ind+0] = nor(0);
				norBuf[3*ind+1] = nor(1);
				norBuf[3*ind+2] = nor(2);
			}
		}
	}
}

template <typename Derived>
void addMatToTriplets(const DenseBase<Derived> &mat, vector<Td> &triplets, int iOff, int jOff){
	for(int i = 0; i < mat.rows(); i++){
		for(int j = 0; j < mat.rows(); j++){
			double val = mat(i,j);
			if(val != 0){
				triplets.push_back(Td(i + iOff, j + jOff, val));
			}
		}
	}
}

void Car::doSusp(int i, int j, int k, int wheelInd, std::set<int> &pressedKeys, std::vector< std::shared_ptr<Collider> > &colliders){
	auto particle = particles[i*numY*numZ + j*numZ + k];
	
	Vector3d up = (particles[(numY-1)*numZ]->x - particles[0]->x).normalized();
	
	Vector3d forward = (particles[0]->x - particles[(numZ-1)]->x).normalized();
	Vector3d right = forward.cross(up);
	up = right.cross(forward);

	std::shared_ptr<Collider> hitting;
	double hitDist = -1;
	
	for(auto collider : colliders) {
		double d = collider->intersect(particle->x, -1 * up);
		if(d >= 0){
			if(hitDist < 0 || d < hitDist){
				hitting = collider;
				hitDist = d;
			}
		}
	}

	double restLength = 0.5;

	if(hitDist > restLength)
		hitDist = -1;

	double length = hitDist < 0 ? restLength : hitDist;

	double rotateDir = 0;
	shared_ptr<Wheel> wheel = wheels[wheelInd];

	if(hitDist >= 0){
		double squish = restLength - hitDist;
	
		// cout << y << " " << restLength << " " << squish << endl;
		Vector3d force = 3 * squish * up;
		f.segment(particle->i, 3) += force;

		glm::vec4 v = glm::vec4(forward(0), forward(1), forward(2), 0);
		glm::vec4 vR = glm::rotate(glm::mat4(1.0f), (float) (-1 * wheel->rotate * M_PI/180), glm::vec3(up(0), up(1), up(2))) * v;
		Vector3d wheelForward = Vector3d(vR.x, vR.y, vR.z);
		
		double drive = 0.25 * 1;

		if(pressedKeys.find(340) != pressedKeys.end()) // SHIFT
			drive *= 5;
		if(pressedKeys.find(341) != pressedKeys.end()) // CTRL
			drive *= 0.25;

		if(pressedKeys.find(87) != pressedKeys.end()){ // W
			f.segment(particle->i, 3) += drive * wheelForward;
		}
		if(pressedKeys.find(83) != pressedKeys.end()){
			f.segment(particle->i, 3) += drive * -1 * wheelForward; // S
		}
	}

	if(pressedKeys.find(81) != pressedKeys.end()){ // Q
		rotateDir = -1;
	}
	if(pressedKeys.find(69) != pressedKeys.end()){ // E
		rotateDir = 1;
	}
	
	if(k == 0 && pressedKeys.find(65) != pressedKeys.end()){ // A
		rotateDir = -1;
		// f.segment(particle->i, 3) += drive * -1 * right;
	}
	if(k == 0 && pressedKeys.find(68) != pressedKeys.end()){ // D
		rotateDir = 1;
		// f.segment(particle->i, 3) += drive * right;
	}

	Vector3d out = right;
	if(wheel->flipped)
		out = -1 * out;
	Vector3d wheelPos = particle->x - up * (length - wheel->radius) + out * 0.25;

	double forwardVel = particle->v.dot(forward);
	double angVel = forwardVel / wheel->radius * 180/M_PI;

	wheel->update(wheelPos, forward, up, rotateDir, angVel);
}

glm::mat4 Car::getMatrix() {
	Vector3d center(0.0, 0.0, 0.0);
	for(int i = 0; i < numX; i++) 
		for(int j = 0; j < numY; j++) 
			for(int k = 0; k < numZ; k++) 
				center += particles[i*numY*numZ + j*numZ + k]->x;
	center /= numX * numY * numZ;

	Vector3d diff = center - lastCenter;
	center = lastCenter + diff / 20;
	lastCenter = center;

	float Bf[16] = {
		1, 0, 0, 0, 
		0, 1, 0, 0, 
		0, 0, 1, 0, 
		0, 0, 0, 1
	};

	glm::mat4 mat = glm::make_mat4(Bf);
	mat[3] = glm::vec4(center(0), center(1), center(2), 1.0f);
	
	return mat;
}

void Car::step(double h, const Vector3d &grav, const vector< shared_ptr<Particle> > spheres, std::set<int> &pressedKeys, bool* keyToggles, std::vector< std::shared_ptr<Collider> > &colliders)
{

	drawMesh = !keyToggles[(unsigned)'v'];
	drawBody = !keyToggles[(unsigned)'b'];

	// for(shared_ptr<Spring> spring : springs)
	// 	spring->E = keyToggles[(unsigned)'j'] ? 4 : 8;

	// M.setZero();
	// K.setZero();
	v.setZero();
	f.setZero();

	vector<Td> Mtrip;
	vector<Td> Ktrip;

	for(shared_ptr<Particle> particle : particles){
		if(!particle->fixed){
			// Matrix3d m;
			// m << particles[i]->m, 0, 0, 
			// 	0, particles[i]->m, 0, 
			// 	0, 0, particles[i]->m;

			// M.block<3,3>(particles[i]->i, particles[i]->i) = m;
			for(int i = 0; i < 3; i++)
				Mtrip.push_back(Td(particle->i + i, particle->i + i, particle->m));

			v.segment(particle->i, 3) = particle->v;
			f.segment(particle->i, 3) = particle->m * grav; // f = mg
		}
	}

	for(shared_ptr<Spring> spring : springs){
		Vector3d dx = spring->p1->x - spring->p0->x;
		double l = dx.norm();
		Vector3d force = spring->E * (l - spring->L) * dx / l;
		
		if(!spring->p0->fixed)
			f.segment(spring->p0->i, 3) += force;
		if(!spring->p1->fixed)
			f.segment(spring->p1->i, 3) -= force;

		Matrix3d a = (1 - (l - spring->L)/l) * (dx * dx.transpose());
		Matrix3d b = (l - spring->L)/l * (dx.transpose() * dx).value() * Matrix3d::Identity();
		Matrix3d ks = spring->E / (l * l) * (a + b);

		// if(!spring->p0->fixed)
		// 	K.block<3,3>(spring->p0->i, spring->p0->i) += -ks;
		// if(!spring->p1->fixed)
		// 	K.block<3,3>(spring->p1->i, spring->p1->i) += -ks;
		// if(!spring->p0->fixed && !spring->p1->fixed){
		// 	K.block<3,3>(spring->p0->i, spring->p1->i) += ks;
		// 	K.block<3,3>(spring->p1->i, spring->p0->i) += ks;
		// }

		if(!spring->p0->fixed)
			addMatToTriplets(-ks, Ktrip, spring->p0->i, spring->p0->i);
		if(!spring->p1->fixed)
			addMatToTriplets(-ks, Ktrip, spring->p1->i, spring->p1->i);
		if(!spring->p0->fixed && !spring->p1->fixed){
			addMatToTriplets(ks, Ktrip, spring->p0->i, spring->p1->i);
			addMatToTriplets(ks, Ktrip, spring->p1->i, spring->p0->i);
		}
	}

	// COLLISIONS

	double c = 20;

	for(shared_ptr<Particle> particle : particles){
		for(shared_ptr<Particle> sphere : spheres){
			Vector3d dx = particle->x - sphere->x;
			double l = dx.norm();
			double d = particle->r + sphere->r - l;
			
			if(d > 0){ // collision
				Vector3d collisionForce = c * d * (dx / l);
				Matrix3d stiffness = c * d * Matrix3d::Identity();

				f.segment(particle->i, 3) += collisionForce;
				// K.block<3,3>(particle->i, particle->i) += stiffness;
				addMatToTriplets(stiffness, Ktrip, particle->i, particle->i);
			}
		}

		double c2 = 20;
		if(particle->x(1) < 0){
			Vector3d dx = Vector3d(0, 1, 0);
			double d = 0 - particle->x(1);
			
			if(d > 0){ // collision
				Vector3d collisionForce = c2 * d * dx;
				Matrix3d stiffness = c2 * d * Matrix3d::Identity();

				f.segment(particle->i, 3) += collisionForce;
				// K.block<3,3>(particle->i, particle->i) += stiffness;
				addMatToTriplets(stiffness, Ktrip, particle->i, particle->i);
			}
		}
	}

	doSusp(0, 0, numZ-1, 0, pressedKeys, colliders);
	doSusp(0, 0, 0, 1, pressedKeys, colliders);
	doSusp(numX-1, 0, numZ-1, 2, pressedKeys, colliders);
	doSusp(numX-1, 0, 0, 3, pressedKeys, colliders);

	// Damping
	double d = 0.001;//0.0015;
	MatrixXd D = d * MatrixXd::Identity(n, n);

	// SOLVE

	// MatrixXd A = M - h * h * K;
	// VectorXd b = M * v + h * f;

	// VectorXd newVel = A.ldlt().solve(b);
	SparseMatrix<double> M(n, n);
	M.setFromTriplets(Mtrip.begin(), Mtrip.end());

	SparseMatrix<double> K(n, n);
	K.setFromTriplets(Ktrip.begin(), Ktrip.end());

	SparseMatrix<double> A = M - h * h * K + h * D;
	VectorXd b = M * v + h * f;

	ConjugateGradient< SparseMatrix<double> > cg;
	cg.setMaxIterations(25);
	cg.setTolerance(1e-6);
	cg.compute(A);
	VectorXd newVel = cg.solveWithGuess(b, v);

	for(int i = 0; i < particles.size(); i++){
		if(!particles[i]->fixed){
			particles[i]->v = newVel.segment(particles[i]->i, 3);
		}
	}

	for(int i = 0; i < particles.size(); i++){
		if(!particles[i]->fixed){
			particles[i]->x += h * particles[i]->v;
		}
	}

	// cout << "\n---" << endl;
	// cout << A << endl;
	// cout << b << endl;
	// cout << newVel << endl;
	
	// Update position and normal buffers
	updatePosNor();

	for(auto wheel : wheels)
		wheel->step(h);
}

void Car::init()
{
	glGenBuffers(1, &posBufID);
	glBindBuffer(GL_ARRAY_BUFFER, posBufID);
	glBufferData(GL_ARRAY_BUFFER, posBuf.size()*sizeof(float), &posBuf[0], GL_DYNAMIC_DRAW);
	
	glGenBuffers(1, &norBufID);
	glBindBuffer(GL_ARRAY_BUFFER, norBufID);
	glBufferData(GL_ARRAY_BUFFER, norBuf.size()*sizeof(float), &norBuf[0], GL_DYNAMIC_DRAW);
	
	// glGenBuffers(1, &texBufID);
	// glBindBuffer(GL_ARRAY_BUFFER, texBufID);
	// glBufferData(GL_ARRAY_BUFFER, texBuf.size()*sizeof(float), &texBuf[0], GL_STATIC_DRAW);
	
	glGenBuffers(1, &eleBufID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, eleBufID);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, eleBuf.size()*sizeof(unsigned int), &eleBuf[0], GL_STATIC_DRAW);
	
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	wheelShape->init();

	set<int> pressedKeys;
	std::vector< std::shared_ptr<Collider> > colliders;
	doSusp(0, 0, numZ-1, 0, pressedKeys, colliders);
	doSusp(0, 0, 0, 1, pressedKeys, colliders);
	doSusp(numX-1, 0, numZ-1, 2, pressedKeys, colliders);
	doSusp(numX-1, 0, 0, 3, pressedKeys, colliders);
	
	assert(glGetError() == GL_NO_ERROR);
}

void Car::draw(shared_ptr<MatrixStack> MV, const shared_ptr<Program> p) const
{
	// // Draw mesh
	glUniform3fv(p->getUniform("kdFront"), 1, Vector3f(1.0, 0.5, 0.0).data());
	// glUniform3fv(p->getUniform("kdFront"), 1, Vector3f(1.0, 0.0, 0.0).data());
	glUniform3fv(p->getUniform("kdBack"),  1, Vector3f(1.0, 1.0, 0.0).data());
	glUniform1f(p->getUniform("alpha"),  1.0);//0.8);
	MV->pushMatrix();
	glUniformMatrix4fv(p->getUniform("MV"), 1, GL_FALSE, glm::value_ptr(MV->topMatrix()));
	
	if(drawBody){
		int h_pos = p->getAttribute("aPos");
		glEnableVertexAttribArray(h_pos);
		glBindBuffer(GL_ARRAY_BUFFER, posBufID);
		glBufferData(GL_ARRAY_BUFFER, posBuf.size()*sizeof(float), &posBuf[0], GL_DYNAMIC_DRAW);
		glVertexAttribPointer(h_pos, 3, GL_FLOAT, GL_FALSE, 0, (const void *)0);
		int h_nor = p->getAttribute("aNor");
		glEnableVertexAttribArray(h_nor);
		glBindBuffer(GL_ARRAY_BUFFER, norBufID);
		glBufferData(GL_ARRAY_BUFFER, norBuf.size()*sizeof(float), &norBuf[0], GL_DYNAMIC_DRAW);
		glVertexAttribPointer(h_nor, 3, GL_FLOAT, GL_FALSE, 0, (const void *)0);
		
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, eleBufID);
		for(int i = 0; i < numX-1; i++) {
			glDrawElements(GL_TRIANGLE_STRIP, 2*numZ, GL_UNSIGNED_INT, (const void *)(2*numZ*i*sizeof(unsigned int)));
			glDrawElements(GL_TRIANGLE_STRIP, 2*numZ, GL_UNSIGNED_INT, (const void *)((2*numZ*(i+numX-1))*sizeof(unsigned int)));

			glDrawElements(GL_TRIANGLE_STRIP, 2*numY, GL_UNSIGNED_INT, (const void *)((2*numY*i + 2*(numX-1)*2*numZ + 2*(numY-1)*2*numZ)*sizeof(unsigned int)));
			glDrawElements(GL_TRIANGLE_STRIP, 2*numY, GL_UNSIGNED_INT, (const void *)((2*numY*(i+numX-1) + 2*(numX-1)*2*numZ + 2*(numY-1)*2*numZ)*sizeof(unsigned int)));
		}
		for(int j = 0; j < numY-1; j++) {
			glDrawElements(GL_TRIANGLE_STRIP, 2*numZ, GL_UNSIGNED_INT, (const void *)((2*numZ*j + 2*(numX-1)*2*numZ)*sizeof(unsigned int)));
			glDrawElements(GL_TRIANGLE_STRIP, 2*numZ, GL_UNSIGNED_INT, (const void *)((2*numZ*(j+numY-1) + 2*(numX-1)*2*numZ)*sizeof(unsigned int)));
		}
		
		glDisableVertexAttribArray(h_nor);
		glDisableVertexAttribArray(h_pos);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	}

	if(drawMesh) {
		glColor3f(1.0f, 0.0f, 0.0f);

		glLineWidth(2.0f);
		glBegin(GL_LINES);

		Vector3d center(0.0, 0.0, 0.0);
		for(int i = 0; i < numX; i++) 
			for(int j = 0; j < numY; j++) 
				for(int k = 0; k < numZ; k++) 
					center += particles[i*numY*numZ + j*numZ + k]->x;
		center /= numX * numY * numZ;

		for(shared_ptr<Spring> spring : springs) {
			float extrude = 0.001;
			Vector3d x1 = spring->p0->x;
			x1 = ((x1-center).norm()+extrude) * (x1-center).normalized() + center;

			Vector3d x2 = spring->p1->x;
			x2 = ((x2-center).norm()+extrude) * (x2-center).normalized() + center;

			glVertex3f(x1(0), x1(1), x1(2));
			glVertex3f(x2(0), x2(1), x2(2));
		}

		glEnd();
	}

	// glLineWidth(8.0f);
	// glBegin(GL_LINES);

	// for(int i = 0; i < numX; i++)  {
	// 	for(int j = 0; j < numY; j++)  {
	// 		for(int k = 0; k < numZ; k++)  {
	// 			if((i == 0 || i == numX-1) || (j == 0 || j == numY-1) || (k == 0 || k == numZ-1)){
	// 				int ind = i*numY*numZ + j*numZ + k;
	// 				Vector3d norm(norBuf[3*ind+0], norBuf[3*ind+1], norBuf[3*ind+2]);

	// 				Vector3d x1 = particles[ind]->x;
	// 				Vector3d x2 = x1 + norm;

	// 				glVertex3f(x1(0), x1(1), x1(2));
	// 				glVertex3f(x2(0), x2(1), x2(2));
	// 			}
	// 		}
	// 	}
	// }

	// glEnd();

	MV->popMatrix();

	for(auto wheel : wheels)
		wheel->draw(MV, p);
        
}
