#pragma once
#ifndef Wheel_H
#define Wheel_H

#include <vector>
#include <set>

#define EIGEN_DONT_ALIGN_STATICALLY
#include <Eigen/Dense>
#include <Eigen/Sparse>

#define GLM_FORCE_RADIANS
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtx/string_cast.hpp>

#include "Shape.h"
#include "MatrixStack.h"
#include "Program.h"

using namespace std;
using namespace Eigen;

#ifndef M_PI
    #define M_PI 3.14159265358979323846
#endif

class Wheel
{

public:

    std::shared_ptr<Shape> wheelShape;

    Eigen::Vector3d position;
    Eigen::Vector3d forward;
    Eigen::Vector3d up;
    double rotate = 0;
    double rotateDir = 0;
    double spinRate = 0;
    double spinAngle = 0;
    bool flipped = false;

    double radius = 0.42;

    Wheel(std::shared_ptr<Shape> _wheelShape, bool _flipped) {
        wheelShape = _wheelShape;
        flipped = _flipped;
    }
	
    void update(Eigen::Vector3d _position, Eigen::Vector3d _forward, Eigen::Vector3d _up, double _rotateDir, double _spinRate) {
        position = _position;
        forward = _forward;
        up = _up;
        rotateDir = _rotateDir;
        spinRate = _spinRate;
    }

    void step(double h) {
        spinAngle += h * spinRate;
        double rotateTarget = rotateDir * 45;
        double step = rotateTarget - rotate;
        double maxStep = 2*90 * h;
        
        if(step > maxStep)
            step = maxStep;
        if(step < -maxStep)
            step = -maxStep;

        rotate += step;
        
        if(rotate > 45)
            rotate = 45;
        if(rotate < -45)
            rotate = -45;
    }

	void draw(std::shared_ptr<MatrixStack> MV, const std::shared_ptr<Program> p) const {
        
        MV->pushMatrix();

        Vector3d left = up.cross(forward);

        float Bf[16] = {
            (float) left(0), (float) left(1), (float) left(2), 0,
            (float) up(0), (float) up(1), (float) up(2), 0, 
            (float) forward(0), (float) forward(1), (float) forward(2), 0, 
            0, 0, 0, 1
        };

        glm::mat4 rotMat = glm::make_mat4(Bf);
        
        MV->translate(position(0), position(1), position(2));
        MV->multMatrix(rotMat);


        // glUniformMatrix4fv(p->getUniform("MV"), 1, GL_FALSE, glm::value_ptr(MV->topMatrix()));
        // glLineWidth(4.0f);
        // glBegin(GL_LINES);

        // glColor3f(1, 0, 0);
        // glVertex3f(0, 0, 0);
        // glVertex3f(1, 0, 0);

        // glColor3f(0, 1, 0);
        // glVertex3f(0, 0, 0);
        // glVertex3f(0, 1, 0);

        // glColor3f(0, 0, 1);
        // glVertex3f(0, 0, 0);
        // glVertex3f(0, 0, 1);

        // glEnd();

        MV->rotate(-rotate * M_PI/180, 0, 1, 0);
        MV->rotate(spinAngle * M_PI/180, 1, 0, 0);

        MV->scale(2.0/30.0);
        MV->rotate(M_PI/2, 0, 1, 0);
        MV->translate(0, 0, 2.75);
        glUniformMatrix4fv(p->getUniform("MV"), 1, GL_FALSE, glm::value_ptr(MV->topMatrix()));
        glUniform3fv(p->getUniform("kdFront"), 1, Vector3f(0.1, 0.1, 0.1).data());
        glUniform3fv(p->getUniform("kdBack"),  1, Vector3f(0.1, 0.1, 0.1).data());
        wheelShape->draw(p);

        MV->popMatrix();

    }

};

#endif